unit MainUnit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls;

const
  AM_PRINTREPORT = WM_USER + $99;

type
  TMainForm = class(TForm)
    TcLabel: TLabel;
    CJCLabel: TLabel;
    SCLabel: TLabel;
    ErrLabel: TLabel;
    TestRadios: TRadioGroup;
    procedure FormCreate(Sender: TObject);
    procedure TestRadiosClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
  protected
    procedure IdleTestConvert(Sender: TObject; var Done: Boolean);
    procedure AMPrintReport(var Msg: TMessage); message AM_PRINTREPORT;
    { Private declarations }
  public
    { Public declarations }
  end;

var
  MainForm: TMainForm;

implementation

{$R *.DFM}

uses
  Math,
  TheseUtils, Temperature, TypeK, LM335;

type
  TPScanVCallback = procedure (pData: PDouble; DataSize: LongWord); stdcall;

function TcC_Convert(pRawData: PWord; RawDataCount: LongWord; pResult: PDouble; ResultCount: LongWord; ExtraGain: Double; pConfig: PByte; ConfigBytes: LongWord): LongWord; stdcall; external 'TcC.dll';
function TcC_StartAcquisition(TcRange, CJCRange, Oversample: Byte; AcqHz: Double; BufferSize, BaseBufferCount: LongWord; pCallback: TPScanVCallback): LongWord; stdcall; external 'TcC.dll';
function TcC_StopAcquisition: LongWord; stdcall; external 'TcC.dll';


function CtFrommV(RangeCode: Byte; mV: TReal): Word;
begin
  mV := 1/1000 * mV;
  if (RangeCode and 4) = 0 then mV := 1/5 * mV;
  if (RangeCode and 2) = 0 then mV := 1/2 * mV;
  if (RangeCode and 1) <> 0 then mV := 1/2 * (mV + 1);
  Result := Min($FFFF, Max($0000, Round(65536 * mV)));
end;

function mVFromCt(RangeCode: Byte; Ct: Word): TReal;
begin
  Result := 1/65536 * Ct;
  if (RangeCode and 1) <> 0 then Result := Result * 2 - 1;
  if (RangeCode and 2) = 0 then Result := Result * 2;
  if (RangeCode and 4) = 0 then Result := Result * 5;
  Result := 1000 * Result;
end;

type
  TErrData = record
    Tot: TReal;
    Wt: Integer;
    MaxErr: TReal;
  end;
  TPErrData = ^TErrData;

var
  TcErr, CJCErr, SCErr: TErrData;
  ErrStatus: array of LongWord;
  LastTest: Integer = 0;
  hErrDataMutex: THandle;

function MyFloatToStr(F: Extended): String;
begin
  Result := FloatToStrF(F, ffFixed, 20, 5);
end;

procedure AddErrStatus(Status: LongWord);
var
  I: Integer;
  
begin
  WaitForSingleObject(hErrDataMutex, INFINITE);
    I := Length(ErrStatus);
    SetLength(ErrStatus, I+1);
    ErrStatus[I] := Status;
  ReleaseMutex(hErrDataMutex);
end;

procedure AddErrDataPt(var ErrData: TErrData; Err: TReal);
begin
  WaitForSingleObject(hErrDataMutex, INFINITE);
    with ErrData do begin
      Tot := Tot + Err;
      Inc(Wt);
      if MaxErr < Err then MaxErr := Err;
    end;
  ReleaseMutex(hErrDataMutex);
end;

function Report(const ErrData: TErrData): String; overload;
begin
  with ErrData do
    Result := MyFloatToStr(Tot / Wt) + ' is ' + IntToStr(Round(100 * Tot / Wt / MaxErr)) + '% of ' + MyFloatToStr(MaxErr)
  ;
end;

procedure TMainForm.AMPrintReport(var Msg: TMessage);
var
  I: Integer;
  Buf: String;

begin
  WaitForSingleObject(hErrDataMutex, INFINITE);
    try
      if TcErr.Wt > 0 then TcLabel.Caption := 'Tc: ' + Report(TcErr);
      if CJCErr.Wt > 0 then CJCLabel.Caption := 'CJC: ' + Report(CJCErr);
      if SCErr.Wt > 0 then SCLabel.Caption := 'SC: ' + Report(SCErr);
      Buf := '';
      for I := 0 to High(ErrStatus) do
        Buf := Buf + (IntToStr(ErrStatus[I]) + ' ')
      ;
      ErrLabel.Caption := Buf;
    except
    end;
  ReleaseMutex(hErrDataMutex);
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  Randomize;

  IsMultiThread := True;

  hErrDataMutex := CreateMutex(nil, False, nil);
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  CloseHandle(hErrDataMutex);
  hErrDataMutex := INVALID_HANDLE_VALUE;
end;

procedure TMainForm.IdleTestConvert(Sender: TObject; var Done: Boolean);
const
  ChCt = 128;

var
  TcC: array of TReal;
  CJCC, CJCmV: TReal;
  iTc, iConfigByte, I: Integer;
  RawData: array of Word;
  TcCResult: array of TReal;
  Status: LongWord;
  Config: array of Byte;
  Err: TReal;
  DummyMsg: TMessage;

begin
  try
    Done := False;

    SetLength(TcC, ChCt);
    for iTc := 0 to ChCt-1 do
      TcC[iTc] := Random * 57.6
    ;
    CJCC := TcC[8];
    CJCmV := TypeK.mVFromF(Temperature.FFromC(CJCC));

    SetLength(Config, 21);
    for iConfigByte := 0 to 15 do
      Config[iConfigByte] := 1
    ;
    Config[$1] := 7;
    Config[$10] := 0;
    Config[$11] := $15;
    Config[$12] := $F0;
    Config[$13] := 0;
    Config[$14] := $70;

    SetLength(RawData, ChCt);
    SetLength(TcCResult, ChCt);
    for iTc := 0 to ChCt-1 do begin
      case iTc of
        8: begin
          RawData[iTc] := CtFrommV(Config[$1], LM335.mVFromC(CJCC));
        end;
        9..15: begin
          RawData[iTc] := CtFrommV(Config[$1], TcC[iTc]);
        end;
        else begin
          RawData[iTc] := CtFrommV(Config[iTc div 8], 100 * (TypeK.mVFromF(Temperature.FFromC(TcC[iTc])) - CJCmV));
        end;
      end;
    end;
    for iTc := 0 to ChCt-1 do
      TcCResult[iTc] := 9001
    ;

    Status := TcC_Convert(@RawData[0], Length(RawData), @TcCResult[0], Length(TcCResult), 100, @Config[0], Length(Config));

    if Status <> ERROR_SUCCESS then begin
      I := Length(ErrStatus);
      SetLength(ErrStatus, I+1);
      ErrStatus[I] := Status;
    end
    else begin
      for iTc := 0 to ChCt-1 do begin
        if TcCResult[iTc] > 9000 then
          Beep
        ;
        Err := Abs(TcCResult[iTc] - TcC[iTc]);
        case iTc of
          8: AddErrDataPt(CJCErr, Err);
          9..15: AddErrDataPt(SCErr, Err);
          else AddErrDataPt(TcErr, Err);
        end;
      end;
    end;

    AMPrintReport(DummyMsg);

  except
    TestRadios.ItemIndex := 0;
    raise;
  end;
end;

procedure ScanVCallback(pData: PDouble; DataSize: LongWord); stdcall;
var
  Data: array of Double;

begin
  SetLength(Data, DataSize div SizeOf(Data[0]));
  Move(pData^, Data[0], Length(Data) * SizeOf(Data[0]));

  AddErrDataPt(TcErr, Data[0]);
  AddErrDataPt(CJCErr, Data[8]);

  PostMessage(MainForm.Handle, AM_PRINTREPORT, 0, 0);
end;

procedure TMainForm.TestRadiosClick(Sender: TObject);
var
  Status: LongWord;
  DummyMsg: TMessage;

begin
  TcErr.Tot := 0;
  TcErr.Wt := 0;
  TcErr.MaxErr := 0;
  CJCErr.Tot := 0;
  CJCErr.Wt := 0;
  CJCErr.MaxErr := 0;
  SCErr.Tot := 0;
  SCErr.Wt := 0;
  SCErr.MaxErr := 0;
  SetLength(ErrStatus, 0);
  TcLabel.Caption := '';
  CJCLabel.Caption := '';
  SCLabel.Caption := '';
  ErrLabel.Caption := '';

  case LastTest of
    1: Application.OnIdle := nil;
    2: begin
      Status := TcC_StopAcquisition;
      if Status <> ERROR_SUCCESS then AddErrStatus(Status);
    end;
  end;

  LastTest := TestRadios.ItemIndex;
  case LastTest of
    1: Application.OnIdle := IdleTestConvert;
    2: begin
      Status := TcC_StartAcquisition(1, 2, 2, 2000, 4*1024, 16, ScanVCallback);
      if Status <> ERROR_SUCCESS then AddErrStatus(Status);
    end;
  end;

  AMPrintReport(DummyMsg);
end;

end.
